"""
"""


import logging
from pathlib import Path
from typing import Optional

import lightning.pytorch as pl
from torch.utils.data import DataLoader

from snp_transformer.config import flatten_nested_dict, load_config, parse_config

std_logger = logging.getLogger(__name__)


def train(config_path: Optional[Path] = None) -> None:
    config_dict = load_config(config_path)
    config = parse_config(config_dict)

    training_cfg = config.training
    training_dataset = config.dataset.training
    validation_dataset = config.dataset.validation
    model = config.model
    logger = training_cfg.trainer.logger  # assumes only one logger
    trainer_kwargs = training_cfg.trainer.to_dict()

    # update config
    flat_config = flatten_nested_dict(config_dict)

    # a hack to get the wandb logger to work
    logger._wandb_init["config"] = flat_config

    # create dataloader:
    train_loader = DataLoader(
        training_dataset,
        batch_size=training_cfg.batch_size,
        shuffle=True,
        collate_fn=model.collate_fn,
        num_workers=training_cfg.num_workers_for_dataloader,
    )
    val_loader = DataLoader(
        validation_dataset,
        batch_size=training_cfg.batch_size,
        shuffle=False,
        collate_fn=model.collate_fn,
        num_workers=training_cfg.num_workers_for_dataloader,
    )

    trainer = pl.Trainer(**trainer_kwargs)

    std_logger.info("Starting training")
    trainer.fit(model=model, train_dataloaders=train_loader, val_dataloaders=val_loader)



import logging
from pathlib import Path

from snp_transformer.train import train

# import os  # noqa
# os.environ["CUDA_LAUNCH_BLOCKING"] = "1"  # noqa


logging.basicConfig(level=logging.INFO)
path = Path(__file__).parent / "config.cfg"

train(path)

# TODO:
#   [x] get it running for 1 full epoch
#   [x] run a model training to convergence
#   fine-tune the model
#     add phenotype prediction
#       [x] add pheno to dataset (is pheno in the dataset? - yes)
#       [x] add pheno to model
#   [x] add positional encoding


# [ ] get training to work using bf16 (current device does not support bf16, but we might get it to work on a different device) or using fp16



from pathlib import Path

from snp_transformer.dataset.loaders import load_pheno, load_pheno_folder


def test_load_pheno(test_data_folder: Path):
    phenos = load_pheno(test_data_folder / "data.pheno")

    assert isinstance(phenos, dict)
    assert len(phenos.items()) == 3
    assert set(phenos.values()) == {0, 1}


def test_load_pheno_folder(test_data_folder: Path):
    phenos = load_pheno_folder(test_data_folder / "phenos")

    assert isinstance(phenos, dict)
    assert isinstance(list(phenos.values())[0], dict)  # noqa
    assert len(phenos.items()) == 2



from pathlib import Path
from typing import Any, Callable

import pandas as pd
import pytest
import torch
from snp_transformer import Individual, IndividualsDataset
from snp_transformer.dataset.loaders import load_details, load_fam, load_sparse
from snp_transformer.model.optimizers import create_adam

TEST_DATA_FOLDER = Path(__file__).parent / "data"


@pytest.fixture()
def individuals(test_data_folder: Path) -> list[Individual]:
    ind_dataset = IndividualsDataset(test_data_folder / "data")

    return [ind_dataset[i] for i in range(len(ind_dataset))]


@pytest.fixture()
def training_dataset(test_data_folder: Path) -> IndividualsDataset:
    return IndividualsDataset(test_data_folder / "data")


@pytest.fixture()
def test_data_folder() -> Path:
    return TEST_DATA_FOLDER


@pytest.fixture()
def fam_path(test_data_folder: Path) -> Path:
    return test_data_folder / "data.fam"


@pytest.fixture()
def sparse_path(test_data_folder: Path) -> Path:
    return test_data_folder / "data.sparse"


@pytest.fixture()
def details_path(test_data_folder: Path) -> Path:
    return test_data_folder / "data.details"


@pytest.fixture()
def details(details_path: Path) -> pd.DataFrame:
    return load_details(details_path)


@pytest.fixture()
def fam(fam_path: Path) -> pd.DataFrame:
    return load_fam(fam_path)


@pytest.fixture()
def sparse(sparse_path: Path) -> pd.DataFrame:
    return load_sparse(sparse_path).to_pandas()


@pytest.fixture()
def optimizer_fn() -> Callable[[Any], torch.optim.Optimizer]:
    return create_adam(lr=1e-3)



from snp_transformer.dataset.dataset import create_individuals_dataset  # noqa: F401
from snp_transformer.model.embedders import create_snp_embedder  # noqa: F401
from snp_transformer.model.model_layers import create_encoder_layer  # noqa: F401
from snp_transformer.model.model_layers import create_transformers_encoder  # noqa: F401
from snp_transformer.model.optimizers import create_adam  # noqa: F401
from snp_transformer.model.positional_embeddings import (  # noqa: F401
    create_absolute_positional_encoding,
    create_tAPE,
)
from snp_transformer.model.task_modules import (
    create_encoder_for_masked_lm,
)  # noqa: F401
from snp_transformer.registry import Registry
from snp_transformer.training.loggers import *  # noqa: F401, F403

from .data_objects import Individual, SNPs
from .dataset.dataset import IndividualsDataset

__all__ = ["IndividualsDataset", "Individual", "SNPs", "Registry"]



from pathlib import Path
from typing import Any, Optional, Union

from lightning.pytorch.loggers.wandb import WandbLogger
from pydantic import BaseModel, ConfigDict

from snp_transformer.dataset.dataset import IndividualsDataset
from snp_transformer.model.task_modules import TrainableModule


class TrainerConfigSchema(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    accelerator: str = "auto"
    strategy: str = "auto"
    devices: Union[list[int], str, int] = "auto"
    num_nodes: int = 1
    precision: str = "32-true"
    logger: WandbLogger
    max_epochs: Optional[int] = None
    min_epochs: Optional[int] = None
    max_steps: int = 10  # use -1 for no limit
    min_steps: Optional[int] = None
    limit_train_batches: Optional[Union[int, float]] = None
    limit_val_batches: Optional[Union[int, float]] = None
    limit_test_batches: Optional[Union[int, float]] = None
    limit_predict_batches: Optional[Union[int, float]] = None
    overfit_batches: Union[int, float] = 0.0
    val_check_interval: Optional[Union[int, float]] = None
    check_val_every_n_epoch: Optional[int] = 1
    num_sanity_val_steps: Optional[int] = None
    log_every_n_steps: Optional[int] = None
    enable_checkpointing: Optional[bool] = None
    enable_progress_bar: Optional[bool] = None
    enable_model_summary: Optional[bool] = None
    accumulate_grad_batches: int = 1
    gradient_clip_val: Optional[Union[int, float]] = None
    gradient_clip_algorithm: Optional[str] = None
    default_root_dir: Optional[Path] = None

    def to_dict(self) -> dict[str, Any]:
        return self.model_dump()


class TrainingConfigSchema(BaseModel):
    batch_size: int
    num_workers_for_dataloader: int = 8
    trainer: TrainerConfigSchema


class DatasetsConfigSchema(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    training: IndividualsDataset
    validation: IndividualsDataset


class ResolvedConfigSchema(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    dataset: DatasetsConfigSchema
    model: TrainableModule
    training: TrainingConfigSchema



import pytest
import torch
from snp_transformer.model.positional_embeddings import AbsolutePositionalEncoding, tAPE
from torch import nn


@pytest.mark.parametrize(
    "module",
    [AbsolutePositionalEncoding(d_model=256), tAPE(d_model=256)],
)
def test_positional_encoding(module: nn.Module):
    x = torch.arange(8_482_506, 8_482_506 + 100)
    x = x.unsqueeze(0)  # batch size 1
    # stack to have a batch size of 2
    x = torch.cat([x, x], dim=0)
    y = module(x)
    assert y.shape == (2, 100, 256)

    # test to and from disk
    torch.save(module, "test.pt")
    module_from_disk = torch.load("test.pt")

    assert isinstance(module_from_disk, nn.Module)



from collections.abc import Sequence
from pathlib import Path
from typing import Any

import pandas as pd
import polars as pl


def read_csv_with_sep_handling(
    path: Path,
    sep: Sequence[str],
    **kwargs: Any,
) -> pd.DataFrame:
    """
    Read in a csv file with multiple possible separators.
    """
    for s in sep:
        try:
            return pd.read_csv(path, sep=s, **kwargs)
        except:  # noqa
            pass
    raise ValueError(f"Could not read file {path} with any of the separators {sep}")


def load_fam(path: Path) -> pd.DataFrame:
    """
    Load in the .fam dataset format.
    """
    fam = read_csv_with_sep_handling(
        path,
        sep=[" ", "\t"],
        header=None,
        names=[
            "fid",
            "iid",
            "father",
            "mother",
            "sex",
            "phenotype",
        ],
        dtype={
            "fid": str,
            "iid": str,
            "father": str,
            "mother": str,
            "sex": int,
            "phenotype": float,
        },
        engine="pyarrow",
    )
    fam = fam.set_index("iid")
    return fam


def load_details(path: Path) -> pd.DataFrame:
    """
    loading in the .details format. It is a custom format on the form:

    ```
    1 8:165998:A:G 0 115998 G A NO_GENE NO_EXON
    1 8:166003:G:A 0 116003 A G NO_GENE NO_EXON
    1 8:166007:C:T 0 116007 T C NO_GENE NO_EXON
    1 8:166009:T:C 0 116009 C T NO_GENE NO_EXON
    ```

    Where the columns are:
        chr snp_id cm bp a1 a2 gene exome
    """
    details = pd.read_csv(
        path,
        sep=" ",
        header=None,
        names=[
            "chr",
            "snp_id",
            "cm",
            "bp",
            "a1",
            "a2",
            "gene",
            "exome",
        ],
        dtype={
            "chr": str,
            "snp_id": str,
            "cm": float,
            "bp": int,
            "a1": str,
            "a2": str,
            "gene": str,
            "exome": str,
        },
        engine="pyarrow",
    )
    return details


def load_bim(path: Path) -> pd.DataFrame:
    """
    Load in the .bim dataset format.
    """
    bim = pd.read_csv(
        path,
        sep="\t",
        header=None,
        names=[
            "chr",
            "snp_id",
            "cm",
            "bp",
            "a1",
            "a2",
        ],
        dtype={
            "chr": str,
            "snp_id": str,
            "cm": float,
            "bp": int,
            "a1": str,
            "a2": str,
        },
    )
    return bim


def convert_bim_to_details(bim: pd.DataFrame) -> pd.DataFrame:
    """
    Convert a .bim dataframe to a .details dataframe.
    """
    details = bim
    details["gene"] = "NO_GENE"
    details["exome"] = "NO_EXON"
    details["snp_id"] = details["snp_id"].astype(str)
    return details


def load_pheno(path: Path) -> dict[str, int]:
    """
    load a phenotype file on the form:

    ```
    1 1 1
    2 2 1
    3 3 0
    ```

    Where the columns are:
        iid fid phenotype
    """

    pheno = pl.read_csv(
        path,
        separator=" ",
        has_header=False,
        dtypes={"iid": pl.Int64, "fid": pl.Int64, "phenotype": pl.Int64},
    )
    return {  # noqa
        iid: phenotype for iid, phenotype in zip(pheno["iid"], pheno["phenotype"])
    }


def load_pheno_folder(path: Path) -> dict[str, dict[str, int]]:
    """
    Assumes a folder on the format is place as path; pheno/*.pheno files one for each phenotype, e.g. pheno/height.pheno.
    """
    assert path.is_dir(), f"Path {path} is not a directory"

    # load all .pheno
    pheno_files = list(path.glob("*.pheno"))
    pheno = {}
    for pheno_file in pheno_files:
        pheno_name = pheno_file.stem
        pheno[pheno_name] = load_pheno(pheno_file)
    return pheno


def load_sparse(path: Path) -> pl.DataFrame:
    """
    Load in the custom .sparse format on the form:

    ```
    Individual SNP Value
    1 1 1
    1 2 1
    1 3 1
    1 4 1
    1 6 1
    2 2 1
    2 9 1
    3 3 2
    ```

    Where the individual ID maps to the .fam file. I.e. the first individual in the sparse file
    is the first individual in the .fam file. So to get their iid, you can do:

        ```
        idx = 1
        fam = load_fam(fam_path)
        offset_by_one = idx - 1  # .sparse is 1-indexed
        iid = fam.iloc[offset_by_one].iid
        ```
    """
    sparse = pl.read_csv(
        path,
        separator=" ",
        dtypes={"Individual": pl.Int64, "SNP": pl.Int64, "Value": pl.Int64},
    )
    return sparse



from functools import partial
from typing import Any, Callable

import torch

from ..registry import Registry


def configure_optimizers(parameters, lr) -> torch.optim.Optimizer:  # noqa: ANN001
    return torch.optim.Adam(parameters, lr=lr)


@Registry.optimizers.register("adam")
def create_adam(lr: float) -> Callable[[Any], torch.optim.Optimizer]:
    return partial(configure_optimizers, lr=lr)



import math
from abc import abstractmethod
from pathlib import Path
from typing import Union

import torch
from snp_transformer.registry import Registry
from torch import Tensor, nn


class PositionalEncodingModule(nn.Module):
    """
    Interface
    """

    d_model: int

    def __init__(self, *args, **kwargs):  # noqa
        super().__init__()

    @abstractmethod
    def forward(self, positions: Tensor) -> Tensor:
        ...

    def to_disk(self, path: Path):
        kwargs = self.kwargs
        model_info = {
            "class_name": type(self).__name__,
            "state_dict": self.state_dict(),
            "kwargs": kwargs,
        }
        torch.save(model_info, path.with_suffix(".pt"))

    @staticmethod
    def from_disk(path: Union[str, Path]) -> "PositionalEncodingModule":
        model_info = torch.load(path)

        class_name = model_info["class_name"]
        kwargs = model_info["kwargs"]

        # Assuming all classes are in the current namespace
        model_class = globals()[class_name]
        model_instance = model_class(**kwargs)
        model_instance.load_state_dict(model_info["state_dict"])

        return model_instance


class AbsolutePositionalEncoding(PositionalEncodingModule):
    def __init__(
        self,
        d_model: int,
        dropout_prob: float = 0.1,
        w_k_constant: float = 10000.0,
    ):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout_prob)
        self.d_model = d_model
        self.w_k_constant = w_k_constant
        self.kwargs = {
            "d_model": d_model,
            "dropout_prob": dropout_prob,
            "w_k_constant": w_k_constant,
        }

    def calculate_pe(self, positions: Tensor) -> Tensor:
        """
        Calculates the positional encoding for the given positions on
        the fly. This is more memory efficient than pre-calculating
        the positional encoding for all positions which is unfeasable
        given the maximum length.
        """
        device = positions.device
        batch_size = positions.size(1)
        div_term = torch.exp(
            torch.arange(0, self.d_model, 2, device=device)
            * (-math.log(self.w_k_constant) / self.d_model),
        )
        pe = torch.zeros(positions.size(0), batch_size, self.d_model, device=device)

        # make the positional broadcastable
        positions = positions.unsqueeze(-1)
        div_term = div_term.view(1, 1, -1)

        pe[:, :, 0::2] = torch.sin(positions * div_term)
        pe[:, :, 1::2] = torch.cos(positions * div_term)
        return pe

    def forward(self, positions: Tensor) -> Tensor:
        """
        Arguments:
            positions: Tensor, shape ``[batch_size, seq_len]``, where ``seq_len`` is the position ids.
        """
        batch_size, seq_len = positions.shape
        # view as batch_size x seq_len -> seq_len x batch_size
        positions = positions.view(seq_len, batch_size)
        pe = self.calculate_pe(positions)
        # view as batch_size x seq_len again
        pe = pe.view(batch_size, seq_len, -1)
        pe = self.dropout(pe)
        return pe


class tAPE(PositionalEncodingModule):
    """
    derived from:
    https://arxiv.org/pdf/2305.16642.pdf
    """

    def __init__(
        self,
        d_model: int,
        dropout_prob: float = 0.1,
        length_sequence: int = 8_482_506,
        w_k_constant: float = 10000.0,
    ):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout_prob)
        self.d_model = d_model
        self.length_sequence = length_sequence
        self.w_k_constant = w_k_constant

        self.kwargs = {
            "d_model": d_model,
            "dropout_prob": dropout_prob,
            "length_sequence": length_sequence,
            "w_k_constant": w_k_constant,
        }

    def calculate_pe(self, positions: Tensor) -> Tensor:
        """
        Calculates the positional encoding for the given positions on
        the fly. This is more memory efficient than pre-calculating
        the positional encoding for all positions which is unfeasable
        given the maximum length.
        """
        device = positions.device
        batch_size = positions.size(1)
        div_term = torch.exp(
            torch.arange(0, self.d_model, 2, device=device)
            * (-math.log(self.w_k_constant) / self.d_model),
        )
        pe = torch.zeros(positions.size(0), batch_size, self.d_model, device=device)
        norm_const = self.d_model / self.length_sequence

        # make the positional broadcastable
        positions = positions.unsqueeze(-1)
        div_term = div_term.view(1, 1, -1)
        pe[:, :, 0::2] = torch.sin((positions * div_term) * norm_const)
        pe[:, :, 1::2] = torch.cos((positions * div_term) * norm_const)
        return pe

    def forward(self, positions: Tensor) -> Tensor:
        """
        Arguments:
            positions: Tensor, shape ``[batch_size, seq_len]``, where ``seq_len`` is the position ids.
        """
        batch_size, seq_len = positions.shape
        # view as batch_size x seq_len -> seq_len x batch_size
        positions = positions.view(seq_len, batch_size)
        pe = self.calculate_pe(positions)
        # view as batch_size x seq_len again
        pe = pe.view(batch_size, seq_len, -1)
        pe = self.dropout(pe)
        return pe


@Registry.embedders.register("absolute_positional_embedding")
def create_absolute_positional_encoding(
    d_model: int,
    dropout_prob: float = 0.1,
    w_k_constant: float = 100_000.0,
) -> AbsolutePositionalEncoding:
    return AbsolutePositionalEncoding(
        d_model=d_model,
        dropout_prob=dropout_prob,
        w_k_constant=w_k_constant,
    )


@Registry.embedders.register("tAPE")
def create_tAPE(
    d_model: int,
    dropout_prob: float = 0.1,
    length_sequence: int = 8_482_506,
    w_k_constant: float = 100_000.0,
) -> tAPE:
    return tAPE(
        d_model=d_model,
        dropout_prob=dropout_prob,
        length_sequence=length_sequence,
        w_k_constant=w_k_constant,
    )



from collections.abc import MutableMapping
from pathlib import Path
from typing import Any, Optional, Union

from confection import Config

from snp_transformer import Registry
from snp_transformer.config.config_schemas import ResolvedConfigSchema

default_config_path = Path(__file__).parent / "default_config.cfg"


def load_config(config_path: Optional[Path] = None) -> Config:
    if config_path is None:
        config_path = default_config_path
    cfg = Config().from_disk(config_path)
    return cfg


def parse_config(config: Config) -> ResolvedConfigSchema:
    resolved = Registry.resolve(config)
    return ResolvedConfigSchema(**resolved)


def flatten_nested_dict(
    d: Union[dict, MutableMapping],
    parent_key: str = "",
    sep: str = ".",
) -> dict:
    """Recursively flatten an infinitely nested config. E.g. {"level1":

    {"level2": "level3": {"level4": 5}}}} becomes:

    {"level1.level2.level3.level4": 5}.

    Args:
        d (Union[Dict, MutableMapping]): Dict to flatten.
        parent_key (str): The parent key for the current dict, e.g. "level1" for the
            first iteration. Defaults to "".
        sep (str): How to separate each level in the dict. Defaults to ".".

    Returns:
        Dict: The flattened dict.
    """

    items: list[tuple[str, Any]] = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten_nested_dict(d=v, parent_key=new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)



from pathlib import Path
from typing import Optional, Union

from lightning.pytorch.loggers.wandb import WandbLogger
from snp_transformer.registry import Registry


@Registry.loggers.register("wandb")
def create_wandb_logger(
    name: Optional[str] = None,
    save_dir: Union[Path, str] = ".",
    version: Optional[str] = None,
    offline: bool = False,
    dir: Optional[Path] = None,  # noqa: A002
    id: Optional[str] = None,  # noqa: A002
    anonymous: Optional[bool] = None,
    project: Optional[str] = None,
    prefix: str = "",
    checkpoint_name: Optional[str] = None,
) -> WandbLogger:
    return WandbLogger(
        name=name,
        save_dir=save_dir,
        version=version,
        offline=offline,
        dir=dir,
        id=id,
        anonymous=anonymous,
        project=project,
        prefix=prefix,
        checkpoint_name=checkpoint_name,
    )



from pathlib import Path

import pytest
from snp_transformer.train import train


@pytest.fixture()
def config_path() -> Path:
    return Path("tests") / "test_configs" / "default_test_config.cfg"


def test_train(config_path: Path):
    """Integration test of the pl trainer"""
    train(config_path)



from snp_transformer.registry import Registry
from torch import nn


@Registry.layers.register("transformer_encoder_layer")
def create_encoder_layer(
    d_model: int,
    nhead: int,
    dim_feedforward: int,
    layer_norm_eps: float = 1e-12,
    norm_first: bool = True,
) -> nn.TransformerEncoderLayer:
    return nn.TransformerEncoderLayer(
        d_model=d_model,
        nhead=nhead,
        dim_feedforward=dim_feedforward,
        layer_norm_eps=layer_norm_eps,
        batch_first=True,
        norm_first=norm_first,
    )


@Registry.layers.register("transformer_encoder")
def create_transformers_encoder(num_layers: int, encoder_layer: nn.Module) -> nn.Module:
    return nn.TransformerEncoder(
        encoder_layer,
        num_layers=num_layers,
        enable_nested_tensor=False,
    )



import pandas as pd
import pytest
from snp_transformer import Individual, SNPs


@pytest.fixture()
def snps(sparse: pd.DataFrame, details: pd.DataFrame) -> SNPs:
    iid = "1"
    ind1 = sparse[sparse["Individual"] == iid]

    snp_values = ind1["Value"].values
    snp_indices = ind1["SNP"].values
    snp_details = details.iloc[snp_indices]  # type: ignore
    snps = SNPs(
        values=list(snp_values),
        chromosomes=snp_details["chr"].values,
        cm=snp_details["cm"].values,
        bp=snp_details["bp"].values,
        a1=snp_details["a1"].values,
        a2=snp_details["a2"].values,
        gene=snp_details["gene"].values,
        exome=snp_details["exome"].values,
    )
    return snps


def test_individuals(fam: pd.DataFrame, snps: SNPs, iid: str = "1"):
    iid = "1"
    ind_fam = fam.loc[iid]

    individual = Individual(
        snps=snps,
        iid=iid,
        fid=ind_fam.fid,
        father=ind_fam.father,
        mother=ind_fam.mother,
        sex=ind_fam.sex,
        phenotype={},
    )

    assert isinstance(individual, Individual)
    assert isinstance(individual.snps, SNPs)
    assert isinstance(individual.phenotype, dict)

    assert len(individual) == len(individual.snps)


def test_snps(
    fam: pd.DataFrame,
    sparse: pd.DataFrame,
    details: pd.DataFrame,
):
    iid = "1"
    ind1 = sparse[sparse["Individual"] == iid]

    snp_values = ind1["Value"].values
    snp_indices = ind1["SNP"].values
    snp_details = details.iloc[snp_indices]  # type: ignore
    ind_fam = fam.loc[iid]
    assert hasattr(ind_fam, "sex")
    snps = SNPs(
        values=list(snp_values),
        chromosomes=snp_details["chr"].values,
        cm=snp_details["cm"].values,
        bp=snp_details["bp"].values,
        a1=snp_details["a1"].values,
        a2=snp_details["a2"].values,
        gene=snp_details["gene"].values,
        exome=snp_details["exome"].values,
    )

    # has length attributes
    assert len(snps) == len(snps.values)



import logging
from abc import abstractmethod
from dataclasses import dataclass

import lightning.pytorch as pl
import torch
from snp_transformer.data_objects import Individual

from ..embedders import Embedder

logger = logging.getLogger(__name__)


@dataclass
class Targets:
    snp_targets: torch.Tensor
    phenotype_targets: torch.Tensor
    is_snp_mask: torch.Tensor
    is_phenotype_mask: torch.Tensor

    def __post_init__(self) -> None:
        self.validate()

    def validate(self) -> None:
        """
        Ensures that not all targets are masked
        """
        pheno_targets = self.phenotype_targets[self.is_phenotype_mask]
        if pheno_targets.numel() != 0 and torch.all(pheno_targets == -1):
            raise ValueError("All phenotype targets are masked")

        snp_targets = self.snp_targets[self.is_snp_mask]
        if torch.all(snp_targets == -1):
            raise ValueError("All SNP targets are masked")


class TrainableModule(pl.LightningModule):
    """
    Interface for a trainable module
    """

    embedding_module: Embedder

    @abstractmethod
    def collate_fn(self, individual: list[Individual]) -> Targets:
        ...



"""
A dataset for loading in patients
"""

import logging
import random
from collections import defaultdict
from pathlib import Path
from typing import Optional

from snp_transformer.data_objects import Individual, SNPs
from snp_transformer.dataset.loaders import (
    load_details,
    load_fam,
    load_pheno_folder,
    load_sparse,
)
from snp_transformer.registry import Registry
from torch.utils.data import Dataset

logger = logging.getLogger(__name__)


class IndividualsDataset(Dataset):
    def __init__(self, path: Path, split_path: Optional[Path] = None) -> None:
        """
        Args:
            path: path to the dataset
            split_path: optional path to a dataframe of individuals to use e.g. for splitting the dataset into train/val/test
        """
        self.path = path
        self.fam_path = path.with_suffix(".fam")
        self.psparse_path = path.with_suffix(".sparse")
        self.details_path = path.with_suffix(".details")
        self.subset_path = split_path
        self.valid_individuals = self._read_split(self.subset_path)

        # ensure that they all exist
        error = f"does not exist in {path}, the following files exist: {list(path.glob('*'))}"
        assert self.fam_path.exists(), f"{self.fam_path} {error}"
        assert self.psparse_path.exists(), f"{self.psparse_path} {error}"
        assert self.details_path.exists(), f"{self.details_path} {error}"

        self.fam = load_fam(self.fam_path)
        self.snp_details = load_details(self.details_path)
        sparse = load_sparse(self.psparse_path)
        self.idx2iid = {i: iid for i, iid in enumerate(self.fam.index.values)}

        self.idx2snp = sparse.partition_by("Individual", as_dict=True)

        pheno_folder = path.parent / "phenos"
        if pheno_folder.exists():
            self.iid2pheno = self.load_phenos(pheno_folder)
        else:
            self.iid2pheno = None
            logger.warning(f"No phenos folder found in {path.parent}")

        self.filter_individuals()

    def filter_individuals(self) -> None:
        if self.valid_individuals is None:
            return
        self.idx2iid = {
            i: iid for i, iid in self.idx2iid.items() if iid in self.valid_individuals
        }

    def _read_split(self, path: Optional[Path]) -> Optional[list[str]]:
        if path is None:
            return None
        with path.open("r") as f:
            iids = f.read().split("\n")
        return iids

    def load_phenos(self, path: Path) -> dict[str, dict[str, int]]:
        pheno2iid = load_pheno_folder(path)
        iid2pheno: dict[str, dict[str, int]] = defaultdict(dict)
        for pheno, iid_map in pheno2iid.items():
            for iid, value in iid_map.items():
                iid2pheno[iid][pheno] = value
        return iid2pheno

    def get_individuals(self) -> list[Individual]:
        return [self[i] for i in range(len(self))]

    def __len__(self) -> int:
        return len(self.idx2iid)

    def get_pheno(self, iid: str) -> dict[str, int]:
        if self.iid2pheno is None:
            return {}
        return self.iid2pheno[iid]

    def __getitem__(self, idx: int) -> Individual:
        iid = self.idx2iid[idx]
        ind = self.idx2snp[idx + 1]  # sparse is 1-indexed
        phenos = self.get_pheno(iid)

        snp_values = ind["Value"].to_numpy()
        snp_indices = ind["SNP"].to_numpy() - 1  # sparse is 1-indexed

        snp_details = self.snp_details.iloc[snp_indices]  # type: ignore
        ind_fam = self.fam.loc[iid]

        snps = SNPs(
            values=list(snp_values),
            chromosomes=list(snp_details["chr"].values),
            cm=list(snp_details["cm"].values),
            bp=list(snp_details["bp"].values),
            a1=list(snp_details["a1"].values),
            a2=list(snp_details["a2"].values),
            gene=list(snp_details["gene"].values),
            exome=list(snp_details["exome"].values),
        )

        individual = Individual(
            snps=snps,
            iid=iid,
            fid=ind_fam.fid,
            father=ind_fam.father,
            mother=ind_fam.mother,
            sex=ind_fam.sex,
            phenotype=phenos,
        )

        return individual

    def train_test_split(
        self,
        prob: float = 0.8,
        train_path_prefix: str = "train",
        test_path_prefix: str = "test",
        seed: int = 42,
    ) -> None:
        """
        Splits the dataset into a train and test set
        """
        random.seed(seed)

        iids = list(self.idx2iid.values())
        random.shuffle(iids)
        split_idx = int(len(iids) * prob)
        train_iids = iids[:split_idx]
        test_iids = iids[split_idx:]

        train_path = self.path.parent / f"{train_path_prefix}.split"
        test_path = self.path.parent / f"{test_path_prefix}.split"

        self._write_split(train_path, train_iids)
        self._write_split(test_path, test_iids)

    def _write_split(self, path: Path, iids: list[str]) -> None:
        with path.open("w") as f:
            f.write("\n".join(iids))


@Registry.datasets.register("individuals_dataset")
def create_individuals_dataset(path: Path) -> IndividualsDataset:
    logger.info("Creating dataset")
    return IndividualsDataset(path)



import pandas as pd
from pydantic import BaseModel, Field


class SNPs(BaseModel):
    """
    The single nucleotide polymorphisms (SNPs) of an individual.
    """

    values: list[int] = Field(..., description="List of SNP values")
    chromosomes: list[str] = Field(
        ...,
        description="Chromosome ids corresponding to the SNPs",
    )
    cm: list[float] = Field(..., description="Genetic distances in centimorgans")
    bp: list[int] = Field(..., description="Base pair coordinates")
    a1: list[str] = Field(..., description="Allele 1")
    a2: list[str] = Field(..., description="Allele 2")
    gene: list[str] = Field(..., description="gene name of the SNP")
    exome: list[str] = Field(..., description="exome name of the SNP")

    def __repr_str__(self, join_str: str) -> str:
        def string_rep(a, v) -> str:  # type: ignore  # noqa
            if isinstance(v, list):
                # return the "a=[1, 2, ...]" representation with the first 2 elements
                return f"{a}=[{v[0]}, {v[1]}, ...]"
            return f"{a}={v!r}"

        return join_str.join(
            repr(v) if a is None else string_rep(a, v) for a, v in self.__repr_args__()
        )

    def __len__(self) -> int:
        return len(self.values)


class Individual(BaseModel):
    """
    A data object representing an individual.
    """

    iid: str = Field(..., description="Individual id")
    fid: str = Field(..., description="Family id")
    father: str = Field(..., description="Father id")
    mother: str = Field(..., description="Mother id")
    sex: int = Field(
        ...,
        description="Sex code ('1' = male, '2' = female, '0' = unknown)",
    )

    phenotype: dict[str, int] = Field(
        ...,
        description="Phenotype of the individual and their value",
    )
    snps: SNPs = Field(..., description="snps of the individual")

    def to_fam(self, include_phenotype: bool = False) -> pd.DataFrame:
        """
        Convert to a pandas DataFrame with the same format as a .fam file.
        """
        fam = pd.DataFrame(
            {
                "fid": self.fid,
                "iid": self.iid,
                "father": self.father,
                "mother": self.mother,
                "sex": self.sex,
            },
        )
        if include_phenotype is False:
            fam["phenotype"] = -9
        else:
            for phenotype, value in self.phenotype.items():
                fam[phenotype] = value

        return fam

    def to_sparse(self) -> pd.DataFrame:
        """
        Convert snps to .sparse format on the following format:

        ```
        Individual SNP Value
        1 1 1
        1 2 1
        1 3 1
        1 4 1
        1 6 1
        ```
        """
        snp_ids: list[int] = list(range(len(self.snps)))
        sparse = pd.DataFrame(
            {
                "Individual": self.iid,
                "SNP": snp_ids,
                "Value": self.snps.values,
            },
        )
        return sparse

    def to_details(self) -> pd.DataFrame:
        """
        Create a .details file on the format

        chr snp_id cm bp a1 a2 gene exome
        1 8:165998:A:G 0 115998 G A NO_GENE NO_EXON
        1 8:166003:G:A 0 116003 A G NO_GENE NO_EXON
        1 8:166007:C:T 0 116007 T C NO_GENE NO_EXON
        1 8:166009:T:C 0 116009 C T NO_GENE NO_EXON
        """

        snp_ids: list[int] = list(range(len(self.snps)))

        details = pd.DataFrame(
            {
                "chr": self.snps.chromosomes,
                "snp_id": snp_ids,
                "cm": self.snps.cm,
                "bp": self.snps.bp,
                "a1": self.snps.a1,
                "a2": self.snps.a2,
                "gene": self.snps.gene,
                "exome": self.snps.exome,
            },
        )
        return details

    def __len__(self) -> int:
        return len(self.snps)



import logging
from pathlib import Path

from snp_transformer.config import load_config, parse_config

logging.basicConfig(level=logging.INFO)
config_path = Path(__file__).parent / "config.cfg"

config_dict = load_config(config_path)
config = parse_config(config_dict)

training_dataset = config.dataset.training

# How many ones vs twos?



"""
minor preprocessing for pretraining for converting bim to details
"""
from pathlib import Path

from snp_transformer.dataset.loaders import convert_bim_to_details, load_bim

home = Path.home()
data_folder = home / "snpher" / "faststorage" / "biobank" / "exomes" / "reformat"

# Convert MHC bim to .details
bim = load_bim(data_folder / "mhc.bim")
details = convert_bim_to_details(bim)
details.to_csv(data_folder / "mhc.details", sep=" ", index=False, header=False)



from pathlib import Path

import pytest
import torch
from snp_transformer.model.embedders import Embedder, Embeddings, InputIds, SNPEmbedder
from snp_transformer.model.positional_embeddings import AbsolutePositionalEncoding


@pytest.mark.parametrize(
    ("embedding_module", "embedding_kwargs"),
    [
        (
            SNPEmbedder,
            {
                "d_model": 32,
                "dropout_prob": 0.1,
                "max_sequence_length": 128,
                "positonal_embedding": AbsolutePositionalEncoding(d_model=32),
            },
        ),
    ],
)
def test_embedding(
    individuals: list,
    embedding_module: Embedder,
    embedding_kwargs: dict,
):
    """
    Test embedding interface
    """
    batch_size = len(individuals)
    embedding_module = embedding_module(**embedding_kwargs)  # type: ignore

    embedding_module.fit(individuals)

    inputs_ids = embedding_module.collate_individuals(individuals)

    assert isinstance(inputs_ids, InputIds)
    assert isinstance(inputs_ids.snp_value_ids, torch.Tensor)
    assert isinstance(inputs_ids.is_padding, torch.Tensor)

    assert inputs_ids.snp_value_ids.shape == inputs_ids.is_padding.shape
    assert inputs_ids.snp_value_ids.shape[0] == batch_size

    max_seq_len_in_batch = inputs_ids.snp_value_ids.shape[1]

    # forward
    outputs = embedding_module(inputs_ids)

    assert isinstance(outputs, Embeddings)

    assert isinstance(outputs.embeddings, torch.Tensor)
    assert isinstance(outputs.is_padding, torch.Tensor)
    assert outputs.embeddings.shape == (
        batch_size,
        max_seq_len_in_batch,
        embedding_kwargs["d_model"],
    )


@pytest.mark.parametrize(
    ("embedding_module", "embedding_kwargs"),
    [
        (
            SNPEmbedder,
            {
                "d_model": 32,
                "dropout_prob": 0.1,
                "max_sequence_length": 128,
                "positonal_embedding": AbsolutePositionalEncoding(d_model=32),
            },
        ),
    ],
)
def test_saving_and_loading(
    embedding_module: Embedder,
    embedding_kwargs: dict,
    individuals: list,
    test_data_folder: Path,
):
    emb = SNPEmbedder(**embedding_kwargs)
    emb.fit(individuals)

    path = test_data_folder / "snp_embedder_checkpoint.pt"

    emb.to_disk(path)

    emb_loaded = embedding_module.from_disk(path)

    assert isinstance(emb_loaded, SNPEmbedder)
    assert emb_loaded.d_model == emb.d_model
    assert emb_loaded.dropout_prob == emb.dropout_prob
    assert emb_loaded.max_sequence_length == emb.max_sequence_length
    assert emb_loaded.is_fitted == emb.is_fitted
    assert emb_loaded.vocab == emb.vocab



import pytest
from snp_transformer import IndividualsDataset
from snp_transformer.model.embedders import SNPEmbedder
from snp_transformer.model.positional_embeddings import AbsolutePositionalEncoding
from snp_transformer.model.task_modules import EncoderForMaskedLM
from snp_transformer.registry import OptimizerFn
from torch import nn
from torch.utils.data import DataLoader

from .conftest import TEST_DATA_FOLDER


def dummy_training_dataset() -> IndividualsDataset:
    return IndividualsDataset(TEST_DATA_FOLDER / "data")


def long_training_dataset() -> IndividualsDataset:
    return IndividualsDataset(TEST_DATA_FOLDER / "long")


@pytest.mark.parametrize(
    "training_dataset",
    [dummy_training_dataset(), long_training_dataset()],
)
def test_model(
    training_dataset: IndividualsDataset,
    optimizer_fn: OptimizerFn,
) -> None:
    # create model:
    d_model = 32
    emb = SNPEmbedder(
        d_model=d_model,
        dropout_prob=0.1,
        max_sequence_length=128,
        positonal_embedding=AbsolutePositionalEncoding(d_model=32),
    )
    encoder_layer = nn.TransformerEncoderLayer(
        d_model=d_model,
        nhead=int(d_model / 4),
        dim_feedforward=d_model * 4,
        batch_first=True,
    )
    encoder = nn.TransformerEncoder(encoder_layer, num_layers=2)

    individuals = [training_dataset[i] for i in range(len(training_dataset))]
    emb.fit(individuals)

    mdl = EncoderForMaskedLM(
        embedding_module=emb,
        encoder_module=encoder,
        create_optimizer_fn=optimizer_fn,
    )

    # create dataloader:
    dataloader = DataLoader(
        training_dataset,
        batch_size=32,
        shuffle=True,
        collate_fn=mdl.collate_fn,
    )

    # run model:
    for input_ids, masked_labels in dataloader:
        output = mdl(input_ids, masked_labels)
        loss = output["loss"]
        loss.backward()  # ensure that the backward pass works



"""
This project uses Invoke (pyinvoke.org) for task management.
Install it via:

```
pip install invoke
```

And then run:

```
inv --list
```

If you do not wish to use invoke you can simply delete this file.
"""


import platform
import re
import shutil
from pathlib import Path
from typing import List, Optional  # noqa

from invoke import Context, Result, task

# Extract supported python versions from the pyproject.toml classifiers key
SUPPORTED_PYTHON_VERSIONS = [
    line.split("::")[-1].strip().replace('"', "").replace(",", "")
    for line in Path("pyproject.toml").read_text().splitlines()
    if "Programming Language :: Python ::" in line
]

NOT_WINDOWS = platform.system() != "Windows"


def echo_header(msg: str):
    print(f"\n--- {msg} ---")


class MsgType:
    # Emojis have to be encoded as bytes to not break the terminal on Windows
    @property
    def DOING(self) -> str:
        return b"\xf0\x9f\xa4\x96".decode() if NOT_WINDOWS else "DOING:"

    @property
    def GOOD(self) -> str:
        return b"\xe2\x9c\x85".decode() if NOT_WINDOWS else "DONE:"

    @property
    def FAIL(self) -> str:
        return b"\xf0\x9f\x9a\xa8".decode() if NOT_WINDOWS else "FAILED:"

    @property
    def WARN(self) -> str:
        return b"\xf0\x9f\x9a\xa7".decode() if NOT_WINDOWS else "WARNING:"

    @property
    def SYNC(self) -> str:
        return b"\xf0\x9f\x9a\x82".decode() if NOT_WINDOWS else "SYNCING:"

    @property
    def PY(self) -> str:
        return b"\xf0\x9f\x90\x8d".decode() if NOT_WINDOWS else ""

    @property
    def CLEAN(self) -> str:
        return b"\xf0\x9f\xa7\xb9".decode() if NOT_WINDOWS else "CLEANING:"

    @property
    def TEST(self) -> str:
        return b"\xf0\x9f\xa7\xaa".decode() if NOT_WINDOWS else "TESTING:"

    @property
    def COMMUNICATE(self) -> str:
        return b"\xf0\x9f\x93\xa3".decode() if NOT_WINDOWS else "COMMUNICATING:"

    @property
    def EXAMINE(self) -> str:
        return b"\xf0\x9f\x94\x8d".decode() if NOT_WINDOWS else "VIEWING:"


msg_type = MsgType()


def git_init(c: Context, branch: str = "main"):
    """Initialize a git repository if it does not exist yet."""
    # If no .git directory exits
    if not Path(".git").exists():
        echo_header(f"{msg_type.DOING} Initializing Git repository")
        c.run(f"git init -b {branch}")
        c.run("git add .")
        c.run("git commit -m 'Init'")
        print(f"{msg_type.GOOD} Git repository initialized")
    else:
        print(f"{msg_type.GOOD} Git repository already initialized")


def setup_venv(
    c: Context,
    python_path: str,
    venv_name: Optional[str] = None,
) -> str:
    """Create a virtual environment if it does not exist yet.

    Args:
        c: The invoke context.
        python_path: The python executable to use.
        venv_name: The name of the virtual environment. Defaults to ".venv".
    """
    if venv_name is None:
        venv_name = ".venv"

    if not Path(venv_name).exists():
        echo_header(
            f"{msg_type.DOING} Creating virtual environment using {msg_type.PY}:{python_path}",
        )
        c.run(f"{python_path} -m venv {venv_name}")
        print(f"{msg_type.GOOD} Virtual environment created")
    else:
        print(f"{msg_type.GOOD} Virtual environment already exists")
    return venv_name


def _add_commit(c: Context, msg: Optional[str] = None):
    print(f"{msg_type.DOING} Adding and committing changes")
    c.run("git add .")

    if msg is None:
        msg = input("Commit message: ")

    c.run(f'git commit -m "{msg}"', pty=NOT_WINDOWS, hide=True)
    print(f"{msg_type.GOOD} Changes added and committed")


def is_uncommitted_changes(c: Context) -> bool:
    git_status_result: Result = c.run(
        "git status --porcelain",
        pty=NOT_WINDOWS,
        hide=True,
    )

    uncommitted_changes = git_status_result.stdout != ""
    return uncommitted_changes


def add_and_commit(c: Context, msg: Optional[str] = None):
    """Add and commit all changes."""
    if is_uncommitted_changes(c):
        uncommitted_changes_descr = c.run(
            "git status --porcelain",
            pty=NOT_WINDOWS,
            hide=True,
        ).stdout

        echo_header(
            f"{msg_type.WARN} Uncommitted changes detected",
        )

        for line in uncommitted_changes_descr.splitlines():
            print(f"    {line.strip()}")
        print("\n")
        _add_commit(c, msg=msg)


def branch_exists_on_remote(c: Context) -> bool:
    branch_name = Path(".git/HEAD").read_text().split("/")[-1].strip()

    branch_exists_result: Result = c.run(
        f"git ls-remote --heads origin {branch_name}",
        hide=True,
    )

    return branch_name in branch_exists_result.stdout


def update_branch(c: Context):
    echo_header(f"{msg_type.SYNC} Syncing branch with remote")

    if not branch_exists_on_remote(c):
        c.run("git push --set-upstream origin HEAD")
    else:
        print("Pulling")
        c.run("git pull")
        print("Pushing")
        c.run("git push")


def create_pr(c: Context):
    c.run(
        "gh pr create --web",
        pty=NOT_WINDOWS,
    )


def update_pr(c: Context):
    echo_header(f"{msg_type.COMMUNICATE} Syncing PR")
    # Get current branch name
    branch_name = Path(".git/HEAD").read_text().split("/")[-1].strip()
    pr_result: Result = c.run(
        "gh pr list --state OPEN",
        pty=False,
        hide=True,
    )

    if branch_name not in pr_result.stdout:
        create_pr(c)
    else:
        open_web = input("Open in browser? [y/n] ")
        if "y" in open_web.lower():
            c.run("gh pr view --web", pty=NOT_WINDOWS)


def exit_if_error_in_stdout(result: Result):
    # Find N remaining using regex

    if "error" in result.stdout:
        errors_remaining = re.findall(r"\d+(?=( remaining))", result.stdout)[
            0
        ]  # testing
        if errors_remaining != "0":
            exit(0)


def pre_commit(c: Context, auto_fix: bool):
    """Run pre-commit checks."""

    # Essential to have a clean working directory before pre-commit to avoid committing
    # heterogenous files under a "style: linting" commit
    if is_uncommitted_changes(c):
        print(
            f"{msg_type.WARN} Your git working directory is not clean. Stash or commit before running pre-commit.",
        )
        exit(1)

    echo_header(f"{msg_type.CLEAN} Running pre-commit checks")
    pre_commit_cmd = "pre-commit run --all-files"
    result = c.run(pre_commit_cmd, pty=NOT_WINDOWS, warn=True)

    exit_if_error_in_stdout(result)

    if ("fixed" in result.stdout or "reformatted" in result.stdout) and auto_fix:
        _add_commit(c, msg="style: Auto-fixes from pre-commit")

        print(f"{msg_type.DOING} Fixed errors, re-running pre-commit checks")
        second_result = c.run(pre_commit_cmd, pty=NOT_WINDOWS, warn=True)
        exit_if_error_in_stdout(second_result)
    else:
        if result.return_code != 0:
            print(f"{msg_type.FAIL} Pre-commit checks failed")
            exit(1)


@task
def static_type_checks(c: Context):
    echo_header(f"{msg_type.CLEAN} Running static type checks")
    c.run("tox -e type", pty=NOT_WINDOWS)


@task
def install(
    c: Context,
    pip_args: str = "",
    msg: bool = True,
    venv_path: Optional[str] = None,
):
    """Install the project in editable mode using pip install"""
    if msg:
        echo_header(f"{msg_type.DOING} Installing project")

    extras = ".[dev,tests,docs]" if NOT_WINDOWS else ".[dev,tests,docs]"
    install_cmd = f"pip install -e {extras} {pip_args}"

    if venv_path is not None and NOT_WINDOWS:
        with c.prefix(f"source {venv_path}/bin/activate"):
            c.run(install_cmd)
            return

    c.run(install_cmd)


def get_python_path(preferred_version: str) -> Optional[str]:
    """Get path to python executable."""
    preferred_version_path = shutil.which(f"python{preferred_version}")

    if preferred_version_path is not None:
        return preferred_version_path

    print(
        f"{msg_type.WARN}: python{preferred_version} not found, continuing with default python version",
    )
    return shutil.which("python")


@task
def setup(c: Context, python_path: Optional[str] = None):
    """Confirm that a git repo exists and setup a virtual environment.

    Args:
        c: Invoke context
        python_path: Path to the python executable to use for the virtual environment. Uses the return value of `which python` if not provided.
    """
    git_init(c)

    if python_path is None:
        # get path to python executable
        python_path = get_python_path(preferred_version="3.9")
        if not python_path:
            print(f"{msg_type.FAIL} Python executable not found")
            exit(1)
    venv_name = setup_venv(c, python_path=python_path)

    install(c, pip_args="--upgrade", msg=False, venv_path=venv_name)

    if venv_name is not None:
        print(
            f"{msg_type.DOING} Activate your virtual environment by running: \n\n\t\t source {venv_name}/bin/activate \n",
        )


@task
def update(c: Context):
    """Update dependencies."""
    echo_header(f"{msg_type.DOING} Updating project")
    install(c, pip_args="--upgrade", msg=False)


@task(iterable="pytest_args")
def test(
    c: Context,
    python_versions: List[str] = (SUPPORTED_PYTHON_VERSIONS[0],),  # noqa # type: ignore
    pytest_args: List[str] = [],  # noqa
):
    """Run tests"""
    # Invoke requires lists as type hints, but does not support lists as default arguments.
    # Hence this super weird type hint and default argument for the python_versions arg.
    echo_header(f"{msg_type.TEST} Running tests")

    python_version_strings = [f"py{v.replace('.', '')}" for v in python_versions]
    python_version_arg_string = ",".join(python_version_strings)

    if not pytest_args:
        pytest_args = [
            "tests",
            "-rfE",
            "--failed-first",
            "-p no:cov",
            "--disable-warnings",
            "-q",
        ]

    pytest_arg_str = " ".join(pytest_args)

    test_result: Result = c.run(
        f"tox -e {python_version_arg_string} -- {pytest_arg_str}",
        warn=True,
        pty=NOT_WINDOWS,
    )

    # If "failed" in the pytest results
    failed_tests = [line for line in test_result.stdout if line.startswith("FAILED")]

    if len(failed_tests) > 0:
        print("\n\n\n")
        echo_header("Failed tests")
        print("\n\n\n")
        echo_header("Failed tests")

        for line in failed_tests:
            # Remove from start of line until /test_
            line_sans_prefix = line[line.find("test_") :]

            # Keep only that after ::
            line_sans_suffix = line_sans_prefix[line_sans_prefix.find("::") + 2 :]
            print(f"FAILED {msg_type.FAIL} #{line_sans_suffix}     ")

    if test_result.return_code != 0:
        exit(test_result.return_code)


def test_for_rej():
    # Get all paths in current directory or subdirectories that end in .rej
    rej_files = list(Path().rglob("*.rej"))

    if len(rej_files) > 0:
        print(f"\n{msg_type.FAIL} Found .rej files leftover from cruft update.\n")
        for file in rej_files:
            print(f"    /{file}")
        print("\nResolve the conflicts and try again. \n")
        exit(1)


@task
def lint(c: Context, auto_fix: bool = False):
    """Lint the project."""
    test_for_rej()
    pre_commit(c=c, auto_fix=auto_fix)
    static_type_checks(c)


@task
def pr(c: Context, auto_fix: bool = False):
    """Run all checks and update the PR."""
    add_and_commit(c)
    lint(c, auto_fix=auto_fix)
    test(c, python_versions=SUPPORTED_PYTHON_VERSIONS)
    update_branch(c)
    update_pr(c)


@task
def docs(c: Context, view: bool = False, view_only: bool = False):
    """
    Build and view docs. If neither build or view are specified, both are run.
    """
    if not view_only:
        echo_header(f"{msg_type.DOING}: Building docs")
        c.run("tox -e docs")

    if view or view_only:
        echo_header(f"{msg_type.EXAMINE}: Opening docs in browser")
        # check the OS and open the docs in the browser
        if platform.system() == "Windows":
            c.run("start docs/_build/html/index.html")
        else:
            c.run("open docs/_build/html/index.html")



from typing import Any, Callable

import catalogue
import torch
from confection import registry

OptimizerFn = Callable[[Any], torch.optim.Optimizer]


class Registry(registry):
    optimizers = catalogue.create("snp_transformers", "optimizers")
    embedders = catalogue.create("snp_transformers", "embedders")

    datasets = catalogue.create("snp_transformers", "datasets")
    tasks = catalogue.create("snp_transformers", "encoders")
    layers = catalogue.create("snp_transformers", "layers")
    loggers = catalogue.create("snp_transformers", "loggers")



import json
import logging
from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Union

import torch
from snp_transformer.data_objects import Individual
from snp_transformer.dataset.dataset import IndividualsDataset
from snp_transformer.registry import Registry
from torch import nn
from torch.nn.utils.rnn import pad_sequence

from .positional_embeddings import PositionalEncodingModule

logger = logging.getLogger(__name__)


@dataclass
class InputIds:
    """

    Attributes:
        type_indexes: Indexes of the domain (e.g. snps and phenotype types such as height, weight)
        value_indexes: Indexes of the values (e.g. snp values, phenotype values)
        is_padding: Boolean tensor indicating which values are padding
    """

    domain_ids: torch.Tensor
    snp_value_ids: torch.Tensor
    snp_position_ids: torch.Tensor
    phenotype_value_ids: torch.Tensor
    phenotype_type_ids: torch.Tensor
    is_padding: torch.Tensor

    def __post_init__(self) -> None:
        self.validate()

    def validate(self) -> None:
        """
        Validated the dataclass. Ensures that phenotype_value_ids are not all -1 and that snp_value_ids are not all -1
        """
        if torch.all(self.phenotype_value_ids == -1):
            raise ValueError("All phenotype values are -1")
        if torch.all(self.snp_value_ids == -1):
            raise ValueError("All snp values are -1")

    def get_batch_size(self) -> int:
        return self.is_padding.shape[0]

    def get_max_seq_len(self) -> int:
        return self.is_padding.shape[1]

    def get_device(self) -> torch.device:
        return self.is_padding.device


@dataclass
class Embeddings:
    embeddings: torch.Tensor
    is_padding: torch.Tensor


@dataclass
class Vocab:
    """

    Attributes:
        snp2idx: Mapping from SNP value to index
        type2idx: Mapping from type to index. Type includes phenotype types (e.g. height, weight) and snp
        phenotype2idx: Mapping from phenotype value to index

    """

    snp2idx: dict[str, int]
    phenotype_value2idx: dict[Union[str, int], int]
    phenotype_type2idx: dict[str, int]
    domain2idx: dict[str, int]
    mask_token: str = "MASK"
    pad_token: str = "PAD"
    snp_token: str = "snp"
    phenotype_token: str = "phenotype"

    @property
    def vocab_size_phenotype_value(self) -> int:
        return len(self.phenotype_value2idx)

    @property
    def vocab_size_snps(self) -> int:
        return len(self.snp2idx)

    @property
    def vocab_size_phenotype_types(self) -> int:
        return len(self.phenotype_type2idx)

    @property
    def vocab_size_types(self) -> int:
        return len(self.domain2idx)

    @property
    def vocab_size(self) -> int:
        # as e.g. snp2idx and phenotype2idx can have overlapping values for the mask and pad tokens
        unique_values = set(
            list(self.domain2idx.values())
            + list(self.snp2idx.values())
            + list(self.phenotype_value2idx.values()),
        )
        return len(unique_values)

    def to_dict(self) -> dict[str, Any]:
        return self.__dict__

    @classmethod
    def from_dict(cls: type["Vocab"], vocab_dict: dict[str, Any]) -> "Vocab":
        return cls(**vocab_dict)


class Embedder(nn.Module):
    """
    Interface for embedding modules
    """

    d_model: int
    vocab: Vocab
    is_fitted: bool

    def __init__(self, *args, **kwargs):  # noqa
        super().__init__()

    @abstractmethod
    def forward(self, *args: Any) -> Embeddings:
        ...

    @abstractmethod
    def fit(self, individuals: list[Individual], *args: Any) -> None:
        ...

    @abstractmethod
    def collate_individuals(self, individual: list[Individual]) -> InputIds:
        ...

    @abstractmethod
    def to_disk(self, save_dir: Path) -> None:
        ...

    @classmethod
    def from_disk(cls: type["Embedder"], save_dir: Path) -> "Embedder":
        ...


class SNPEmbedder(Embedder):
    mask_token = "MASK"
    pad_token = "PAD"

    def __init__(
        self,
        d_model: int,
        dropout_prob: float,
        max_sequence_length: int,
        positonal_embedding: PositionalEncodingModule,
    ):
        super().__init__()

        self.d_model = d_model
        self.max_sequence_length = max_sequence_length
        self.dropout_prob = dropout_prob

        assert (
            d_model == positonal_embedding.d_model
        ), "d_model is not the same for positional embedding and embedder"
        self.positional_embedding = positonal_embedding

        self.is_fitted: bool = False

        self.LayerNorm = nn.LayerNorm(d_model, eps=1e-12)
        self.dropout = nn.Dropout(dropout_prob)

    def check_if_fitted(self) -> None:
        if not self.is_fitted:
            raise RuntimeError("Model must be fitted before use")

    def initialize_embeddings_layers(
        self,
        vocab: Vocab,
    ) -> None:
        self.vocab = vocab

        self.snp_embedding = nn.Embedding(
            num_embeddings=vocab.vocab_size_snps,
            embedding_dim=self.d_model,
            padding_idx=vocab.snp2idx[self.pad_token],
        )

        self.phenotype_value_embedding = nn.Embedding(
            num_embeddings=vocab.vocab_size_phenotype_value,
            embedding_dim=self.d_model,
            padding_idx=vocab.phenotype_value2idx[self.pad_token],
        )

        self.phenotype_type_embedding = nn.Embedding(
            num_embeddings=vocab.vocab_size_phenotype_types,
            embedding_dim=self.d_model,
            padding_idx=vocab.phenotype_type2idx[self.pad_token],
        )

        self.domain_embedding = nn.Embedding(
            num_embeddings=vocab.vocab_size_types,
            embedding_dim=self.d_model,
            padding_idx=vocab.domain2idx[self.pad_token],
        )

        self.is_fitted = True

    def forward(self, inputs: InputIds) -> Embeddings:
        self.check_if_fitted()

        batch_size = inputs.get_batch_size()
        max_seq_len = inputs.get_max_seq_len()

        device = inputs.get_device()

        # start embeddings as a zero tensor
        embeddings = torch.zeros(
            (batch_size, max_seq_len, self.d_model),
            dtype=torch.float,
            device=device,
        )

        # add the domain (e.g. snps, phenotype) embeddings
        embeddings += self.domain_embedding(inputs.domain_ids)
        # NOTE: that the embeddings are added but that tokens for different domains
        # are not influenced by eachother as for the respective tokens we simply
        # add the zero tensor
        embeddings += self.snp_embedding(inputs.snp_value_ids)
        # add the phenotype value embeddings only to the phenotype embeddings
        embeddings += self.phenotype_value_embedding(inputs.phenotype_value_ids)
        # add the phenotype type embeddings only to the phenotype embeddings
        embeddings += self.phenotype_type_embedding(inputs.phenotype_type_ids)

        embeddings += self.positional_embedding(inputs.snp_position_ids)

        embeddings = self.LayerNorm(embeddings)
        embeddings = self.dropout(embeddings)

        return Embeddings(
            embeddings=embeddings,
            is_padding=inputs.is_padding,
        )

    def _collate_individual(self, individual: Individual) -> dict[str, torch.Tensor]:
        """
        Collate an indivivual into ids
        """
        # SNP Values
        snps = individual.snps
        snp_ids = torch.tensor(snps.values, dtype=torch.long)
        snp_pos_ids = torch.tensor(snps.bp, dtype=torch.long)

        # Phenotypes Values
        phenotypes = list(individual.phenotype.values())
        phenotype_ids = torch.tensor(phenotypes, dtype=torch.long)

        # Add padding for SNPs and Phenotypes
        # we want a vector on the length of the concatenated snps and phenotypes
        # with the padding tokens at the beginning for snps
        # and at the end for phenotypes
        snp_padding_id = self.vocab.snp2idx[self.pad_token]
        phenotype_padding_id = self.vocab.phenotype_value2idx[self.pad_token]

        snp_padding = torch.ones_like(phenotype_ids, dtype=torch.long) * snp_padding_id
        phenotype_padding = (
            torch.ones_like(snp_ids, dtype=torch.long) * phenotype_padding_id
        )
        snp_ids = torch.cat([snp_padding, snp_ids])
        phenotype_ids = torch.cat([phenotype_ids, phenotype_padding])

        # Phenotype Types
        phenotype_types_ = [
            self.vocab.domain2idx[ptype] for ptype in individual.phenotype
        ]
        phenotype_types = torch.tensor(phenotype_types_, dtype=torch.long)

        phenotype_type_padding_id = self.vocab.phenotype_type2idx[self.pad_token]
        phenotype_type_padding = (
            torch.ones_like(snp_ids, dtype=torch.long) * phenotype_type_padding_id
        )
        phenotype_types = torch.cat([phenotype_types, phenotype_type_padding])

        # Domain
        snp_idx = self.vocab.domain2idx["snp"]
        pheno_idx = self.vocab.domain2idx["phenotype"]
        domain_ids_: list[int] = [pheno_idx] * len(phenotypes) + [snp_idx] * len(snps)
        domain_ids = torch.tensor(domain_ids_, dtype=torch.long)

        # Padding (for masking)
        is_padding = torch.zeros_like(domain_ids, dtype=torch.bool)

        inputs_ids = {
            "snp_value_ids": snp_ids,
            "snp_position_ids": snp_pos_ids,
            "phenotype_value_ids": phenotype_ids,
            "phenotype_type_ids": phenotype_types,
            "domain_ids": domain_ids,
            "is_padding": is_padding,
        }
        return inputs_ids

    def collate_individuals(self, individual: list[Individual]) -> InputIds:
        """
        Handles padding and indexing by converting each to an index tensor
        """
        self.check_if_fitted()

        inds: list[dict[str, torch.Tensor]] = [
            self._collate_individual(ind) for ind in individual
        ]
        padded_seqs = self._pad_sequences(inds)
        return InputIds(
            domain_ids=padded_seqs["domain_ids"],
            snp_value_ids=padded_seqs["snp_value_ids"],
            snp_position_ids=padded_seqs["snp_position_ids"],
            phenotype_value_ids=padded_seqs["phenotype_value_ids"],
            phenotype_type_ids=padded_seqs["phenotype_type_ids"],
            is_padding=padded_seqs["is_padding"],
        )

    def _pad_sequences(
        self,
        sequences: list[dict[str, torch.Tensor]],
    ) -> dict[str, torch.Tensor]:
        keys = list(sequences[0].keys())
        any_key = keys[0]

        max_seq_len = max([len(p[any_key]) for p in sequences])
        if max_seq_len <= self.max_sequence_length:
            logger.warning(
                "Sequence longer than max sequence length, truncating to max length",
            )
            max_seq_len = self.max_sequence_length

        output = {}
        for key in keys:
            if key == "is_padding":
                continue
            values = [p[key][:max_seq_len] for p in sequences]
            values_w_padding = pad_sequence(
                values,
                batch_first=True,
                padding_value=self.vocab.domain2idx[self.pad_token],
            )
            output[key] = values_w_padding

        is_padding = [p["is_padding"][:max_seq_len] for p in sequences]
        is_padding_w_padding = pad_sequence(
            is_padding,
            batch_first=True,
            padding_value=True,
        )
        output["is_padding"] = is_padding_w_padding
        return output

    def fit(
        self,
        individuals: list[Individual],
    ) -> None:
        # could also be estimated from the data but there is no need for that
        snp2idx = {self.pad_token: 0, self.mask_token: 1, "1": 2, "2": 3}

        domain2idx: dict[str, int] = {
            self.pad_token: 0,
            "snp": 1,
            "phenotype": 2,
        }
        phenotype_value2idx: dict[Union[int, str], int] = {
            self.pad_token: 0,
            self.mask_token: 1,
        }
        phenotype_type2idx: dict[str, int] = {
            self.pad_token: 0,
        }

        for ind in individuals:
            for pheno_type, value in ind.phenotype.items():
                if pheno_type not in phenotype_type2idx:
                    phenotype_type2idx[pheno_type] = len(phenotype_type2idx)
                if value not in phenotype_value2idx:
                    phenotype_value2idx[value] = len(phenotype_value2idx)

        vocab: Vocab = Vocab(
            snp2idx=snp2idx,
            domain2idx=domain2idx,
            phenotype_value2idx=phenotype_value2idx,
            phenotype_type2idx=phenotype_type2idx,
            snp_token="snp",
            phenotype_token="phenotype",
        )

        self.initialize_embeddings_layers(vocab)

    def to_disk(self, save_dir: Path):
        """
        Save the (trained) embedding to disk
        """
        save_dir.mkdir(exist_ok=True, parents=True)

        kwargs = {
            "d_model": self.d_model,
            "dropout_prob": self.dropout_prob,
            "max_sequence_length": self.max_sequence_length,
        }

        config_path = save_dir / "embedder_config.json"
        with config_path.open("w") as fp:
            json.dump(kwargs, fp)

        vocab_path = save_dir / "vocab.json"
        with vocab_path.open("w") as fp:
            json.dump(self.vocab.to_dict(), fp)

        # save positonal embedding
        self.positional_embedding.to_disk(save_dir / "positional_embedding.pt")
        torch.save(self.state_dict(), save_dir / "embedder.pt")

    @classmethod
    def from_disk(cls: type["SNPEmbedder"], save_dir: Path) -> "SNPEmbedder":
        """
        Load the (trained) embedding from disk
        """
        assert save_dir.is_dir()

        config_path = save_dir / "embedder_config.json"
        with config_path.open() as fp:
            kwargs = json.load(fp)

        vocab_path = save_dir / "vocab.json"
        with vocab_path.open() as fp:
            vocab_dict = json.load(fp)

        vocab = Vocab.from_dict(vocab_dict)

        kwargs["positonal_embedding"] = PositionalEncodingModule.from_disk(
            save_dir / "positional_embedding.pt",
        )

        embedder = cls(**kwargs)
        embedder.initialize_embeddings_layers(vocab)
        embedder.load_state_dict(torch.load(save_dir / "embedder.pt"))

        return embedder


@Registry.embedders.register("snp_embedder")
def create_snp_embedder(
    d_model: int,
    dropout_prob: float,
    max_sequence_length: int,
    positional_embedding: PositionalEncodingModule,
    individuals: Union[list[Individual], IndividualsDataset],
    checkpoint_path: Union[Path, None] = None,
) -> Embedder:
    should_load_ckpt = (
        checkpoint_path is not None
        and checkpoint_path.is_dir()
        and list(checkpoint_path.iterdir())  # is not empty
    )

    if should_load_ckpt:
        emb = SNPEmbedder.from_disk(checkpoint_path)  # type: ignore

        kwargs_match = (
            emb.d_model == d_model
            and emb.dropout_prob == dropout_prob
            and emb.max_sequence_length == max_sequence_length
        )
        if kwargs_match:
            logger.info(f"Loaded embedder from {checkpoint_path}")
            return emb
        logger.warning(
            "Embedder kwargs do not match checkpoint kwargs, ignoring checkpoint",
        )

    emb = SNPEmbedder(
        d_model=d_model,
        dropout_prob=dropout_prob,
        max_sequence_length=max_sequence_length,
        positonal_embedding=positional_embedding,
    )

    if isinstance(individuals, IndividualsDataset):
        individuals = individuals.get_individuals()
    emb.fit(individuals)

    if checkpoint_path is not None:
        emb.to_disk(checkpoint_path)

    return emb



"""
Time taken with
- naive pandas implementation:
    - 1:17:8
- Apache Arrow
    - 1:12:39
- With conversion to psparse (polars and pandas)
    - it takes aobut ____
    - but requries a preprocessing which performs ~200-300 individuals pr. second (a total of 200634 individuals)
- with just polars for .sparse
    - 0:0:11

Then there is no need to use the psparse format
"""

import datetime
from pathlib import Path

from snp_transformer.dataset.dataset import IndividualsDataset

st = datetime.datetime.now()

home = Path.home()
data_folder = home / "snpher" / "faststorage" / "biobank" / "exomes" / "reformat"
data_path = data_folder / "mhc"
ds = IndividualsDataset(data_path)
et = datetime.datetime.now()
time_taken = et - st

print(
    f"Time taken to load dataset: {time_taken.seconds//3600}:{(time_taken.seconds//60)%60}:{time_taken.seconds%60}",
)



from pathlib import Path

from snp_transformer import Individual, IndividualsDataset


def test_individuals_dataset(test_data_folder: Path):
    ind_dataset = IndividualsDataset(test_data_folder / "data")

    assert len(ind_dataset) == 3
    ind = ind_dataset[0]
    assert isinstance(ind, Individual)


def test_individuals_dataset_w_split(test_data_folder: Path):
    ind_dataset = IndividualsDataset(
        test_data_folder / "data",
        split_path=test_data_folder / "data_train.split",
    )

    assert len(ind_dataset) == 2
    ind = ind_dataset[0]
    assert isinstance(ind, Individual)


def test_individuals_dataset_train_test_split(test_data_folder: Path):
    ind_dataset = IndividualsDataset(test_data_folder / "data")
    ind_dataset.train_test_split(
        prob=0.8,
        train_path_prefix="tmp_train",
        test_path_prefix="tmp_test",
        seed=42,
    )

    assert len(ind_dataset) == 3, "The original dataset should not be changed"
    train_split_path = test_data_folder / "tmp_train.split"
    test_split_path = test_data_folder / "tmp_test.split"
    assert train_split_path.exists()
    assert test_split_path.exists()

    with train_split_path.open() as f:
        train_iids = f.read().split("\n")
    with test_split_path.open() as f:
        test_iids = f.read().split("\n")

    assert len(train_iids) == 2
    assert len(test_iids) == 1

    # clean up
    train_split_path.unlink()
    test_split_path.unlink()



import logging
import os
from pathlib import Path

from snp_transformer.train import train

logging.basicConfig(level=logging.INFO)
path = Path(__file__).parent / "config.cfg"

os.environ["CUDA_LAUNCH_BLOCKING"] = "1"

print("created obj on cuda")
train(path)


# Test på en anden GPU!
# hmm der er måske noget med indexing af phenotype_value_ids som er -1?



import logging
from copy import copy
from typing import Literal, Union

import torch
from torch import nn
from torchmetrics.classification import MulticlassAccuracy

from ...registry import OptimizerFn, Registry
from ..embedders import Embedder, InputIds, Vocab
from .trainable_modules import Targets, TrainableModule

logger = logging.getLogger(__name__)


class EncoderForMaskedLM(TrainableModule):
    """A LM head wrapper for masked language modeling."""

    ignore_index = -1

    def __init__(
        self,
        embedding_module: Embedder,
        encoder_module: nn.TransformerEncoder,
        create_optimizer_fn: OptimizerFn,
        mask_phenotype: bool = True,
    ):
        super().__init__()
        self.mask_phenotype = mask_phenotype
        self.save_hyperparameters(ignore=["encoder_module", "embedding_module"])
        self.initialize_model(embedding_module, encoder_module)

        self.loss = nn.CrossEntropyLoss(ignore_index=self.ignore_index)
        self.create_optimizer_fn = create_optimizer_fn
        self.initialize_metrics()

    def initialize_metrics(self):
        vocab: Vocab = self.embedding_module.vocab
        self.accuracy_pheno = MulticlassAccuracy(
            num_classes=vocab.vocab_size_phenotype_value,
            ignore_index=self.ignore_index,
        )
        self.accuracy_snp = MulticlassAccuracy(
            num_classes=vocab.vocab_size_snps,
            ignore_index=self.ignore_index,
        )

    def initialize_model(
        self,
        embedding_module: Embedder,
        encoder_module: nn.TransformerEncoder,
    ) -> None:
        self.embedding_module = embedding_module
        self.encoder_module = encoder_module

        self.d_model = self.embedding_module.d_model
        vocab: Vocab = self.embedding_module.vocab

        self.snp_head = nn.Linear(self.d_model, vocab.vocab_size_snps)

        if self.mask_phenotype:
            self.phenotype_head = nn.Linear(
                self.d_model,
                vocab.vocab_size_phenotype_value,
            )

        self.loss = nn.CrossEntropyLoss(ignore_index=-1)

    @property
    def device(self) -> torch.device:
        return next(self.parameters()).device

    def forward(
        self,
        inputs: InputIds,
        targets: Targets,
    ) -> dict[str, Union[torch.Tensor, dict[str, torch.Tensor]]]:
        embeddings = self.embedding_module(inputs)

        encoded_individuals = self.encoder_module(
            embeddings.embeddings,
            src_key_padding_mask=embeddings.is_padding,
        )

        logits_snp = self.snp_head(encoded_individuals)

        loss_snp = self.loss(
            logits_snp.view(-1, logits_snp.size(-1)),
            targets.snp_targets.view(-1),
        )

        snp_preds = torch.argmax(logits_snp, dim=-1)
        snp_acc = self.accuracy_snp(snp_preds, targets.snp_targets)

        result = {
            "loss": loss_snp,
            "SNP Loss": loss_snp,
            "SNP Accuracy": snp_acc,
        }

        if self.mask_phenotype:
            logits_pheno = self.phenotype_head(encoded_individuals)
            loss_pheno = self.loss(
                logits_pheno.view(-1, logits_pheno.size(-1)),
                targets.phenotype_targets.view(-1),
            )
            pheno_preds = torch.argmax(logits_pheno, dim=-1)
            pheno_acc = self.accuracy_pheno(pheno_preds, targets.phenotype_targets)

            # check if loss is nan (this happens when there are no phenotype values)
            if torch.isnan(loss_pheno):
                if torch.all(targets.is_phenotype_mask == False):  # noqa: E712
                    loss_pheno = torch.tensor(0.0, device=self.device)
                else:
                    raise ValueError("Pheno loss is nan")
            result["loss"] += loss_pheno  # assumes equal weighting of domains
            result["Phenotype Loss"] = loss_pheno
            result["Phenotype Accuracy"] = pheno_acc

        # check if loss is nan (this happens when there are no snps)
        if torch.isnan(result["loss"]):
            raise ValueError("Loss is nan")

        return result

    def mask(
        self,
        domain_ids_tensor: torch.Tensor,
        domain_vocab_size: int,
        mask_token_id: int,
        is_padding: torch.Tensor,
        masking_prob: float = 0.15,
        replace_with_mask_prob: float = 0.8,
        replace_with_random_prob: float = 0.1,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Masking function for the task
        """
        masked_lm_labels = domain_ids_tensor.clone()

        # Mask 15 % of the tokens
        prob = torch.rand(domain_ids_tensor.shape)
        mask = prob < masking_prob

        masked_lm_labels[~mask] = -1  # -1 will be ignored in loss function

        prob /= masking_prob

        # 80% of the time, replace with [MASK] token
        mask[mask.clone()] = prob[mask] < replace_with_mask_prob
        domain_ids_tensor[mask] = mask_token_id

        # 10% of the time, replace with random token
        prob /= 0.8
        mask[mask.clone()] = prob[mask] < replace_with_random_prob

        domain_ids_tensor[mask] = torch.randint(
            0,
            domain_vocab_size,
            mask.sum().shape,
        )

        # -> rest 10% of the time, keep the original word

        # ensure that at least one token is masked to avoid a nan loss
        no_pad_labels = masked_lm_labels[~is_padding]
        if torch.all(no_pad_labels == -1):
            # mask the first token (safe but not random)
            i = 0
            masked_lm_labels[i] = domain_ids_tensor[i]
            domain_ids_tensor[i] = mask_token_id

        return domain_ids_tensor, masked_lm_labels

    def masking_fn(
        self,
        padded_sequence_ids: InputIds,
    ) -> tuple[InputIds, Targets]:
        vocab = self.embedding_module.vocab
        is_padding = padded_sequence_ids.is_padding

        snp_mask_token_id = vocab.snp2idx[vocab.mask_token]
        snp_value_ids = copy(padded_sequence_ids.snp_value_ids)

        is_snp_mask = (
            padded_sequence_ids.domain_ids == vocab.domain2idx[vocab.snp_token]
        )

        snp_ids_masked, snp_masked_label = self.mask(
            snp_value_ids,
            domain_vocab_size=vocab.vocab_size_snps,
            mask_token_id=snp_mask_token_id,
            is_padding=is_padding,
        )

        if self.mask_phenotype:
            phenotype_mask_token_id = vocab.phenotype_value2idx[vocab.mask_token]
            pheno_value_ids = copy(padded_sequence_ids.phenotype_value_ids)
            pheno_ids_masked, pheno_masked_label = self.mask(
                pheno_value_ids,
                domain_vocab_size=vocab.vocab_size_phenotype_value,
                mask_token_id=phenotype_mask_token_id,
                is_padding=is_padding,
            )

        else:
            pheno_ids_masked = torch.clone(padded_sequence_ids.phenotype_value_ids)
            pheno_masked_label = torch.clone(padded_sequence_ids.phenotype_value_ids)

        # Make sure to ignore padding when calculating loss and token from other domains
        is_pheno_or_padding = ~is_snp_mask | is_padding
        is_snp_or_padding = is_snp_mask | is_padding
        snp_masked_label[is_padding] = self.ignore_index
        pheno_masked_label[is_snp_or_padding] = self.ignore_index

        masked_sequence_ids = InputIds(
            domain_ids=padded_sequence_ids.domain_ids,
            snp_value_ids=snp_ids_masked,
            snp_position_ids=padded_sequence_ids.snp_position_ids,
            phenotype_value_ids=pheno_ids_masked,
            phenotype_type_ids=padded_sequence_ids.phenotype_type_ids,
            is_padding=is_padding,
        )

        targets = Targets(
            snp_targets=snp_masked_label,
            phenotype_targets=pheno_masked_label,
            is_snp_mask=~is_pheno_or_padding,
            is_phenotype_mask=~is_snp_or_padding,
        )

        return masked_sequence_ids, targets

    def collate_fn(self, individuals: list) -> tuple[InputIds, Targets]:
        """
        Takes a list of individuals and returns a dictionary of padded sequence ids.
        """
        padded_sequence_ids = self.embedding_module.collate_individuals(individuals)
        masked_sequence_ids, masked_labels = self.masking_fn(padded_sequence_ids)
        return masked_sequence_ids, masked_labels

    def training_step(
        self,
        batch: tuple[InputIds, Targets],
        batch_idx: int,  # noqa: ARG002
    ) -> Union[torch.Tensor, dict[str, torch.Tensor]]:
        x, y = batch
        output = self.forward(x, y)
        self.log_step(output, batch_size=x.get_batch_size(), mode="Training")
        return output["loss"]

    def validation_step(
        self,
        batch: tuple[InputIds, Targets],
        batch_idx: int,  # noqa: ARG002
    ) -> Union[torch.Tensor, dict[str, torch.Tensor]]:
        x, y = batch
        output = self.forward(x, y)

        self.log_step(output, x.get_batch_size(), mode="Validation")
        return output["loss"]

    def log_step(
        self,
        output: dict,
        batch_size: int,
        mode: Literal["Validation", "Training"],
    ) -> None:
        log_kwargs = {"batch_size": batch_size, "sync_dist": True}
        for key in output:
            self.log(f"{mode} {key}", output[key], **log_kwargs)

    def configure_optimizers(self) -> torch.optim.Optimizer:
        return self.create_optimizer_fn(self.parameters())


@Registry.tasks.register("masked_lm")
def create_encoder_for_masked_lm(
    embedding_module: Embedder,
    encoder_module: nn.TransformerEncoder,
    create_optimizer_fn: OptimizerFn,
    mask_phenotype: bool,
) -> EncoderForMaskedLM:
    logger.info("Creating task module for masked lm")
    return EncoderForMaskedLM(
        embedding_module=embedding_module,
        encoder_module=encoder_module,
        create_optimizer_fn=create_optimizer_fn,
        mask_phenotype=mask_phenotype,
    )
